import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions
import time
from lal import GreenwichMeanSiderealTime
from astropy.time import Time
from astropy import coordinates as coord
import corner
import os
import shutil
import h5py
import json
import sys
from universal_divergence import estimate
import pandas as pd
import bilby
from bilby.core.prior import Uniform

class DataGenerator(tf.keras.utils.Sequence):
    'Generates data for Keras'
    def __init__(self, input_dir, samples_per_file=1000, batch_size=512, x_dim=15, y_dim=1024, n_channels=3, shuffle=True):
        'Initialization'
        self.x_dim = x_dim
        self.y_dim = y_dim
        self.batch_size = batch_size
        self.fvec = tf.range(params['ndata']/2 + 1)/params['duration'] 
        self.samples_per_file = samples_per_file
        self.n_channels = n_channels
        self.shuffle = shuffle
        self.filenames = [os.path.join(input_dir,f) for f in os.listdir(input_dir)]       
        self.on_epoch_end()

    def __len__(self):
        'Denotes the number of batches per epoch'
        return min(40,int(np.floor(len(self.filenames)*self.samples_per_file / self.batch_size)))

    def __getitem__(self, index):
        'Generate one batch of data'
        # Generate indexes of the batch
        #indexes = self.indexes[index*self.batch_size:(index+1)*self.batch_size]

        # Find list of IDs
        #list_IDs_temp = [self.list_IDs[k] for k in indexes]

        # Generate data
        x, y = self.__data_generation(self.filenames[index])

        return x, y

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        self.indexes = np.arange(len(self.filenames))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)
 
    def __data_generation(self,filename):

        idx = np.sort(np.random.choice(self.samples_per_file,self.batch_size,replace=False))
        x = np.reshape(np.array(h5py.File(filename, 'r')['x_data'][idx]),(self.batch_size,self.x_dim))
        y = np.transpose(np.reshape(np.array(h5py.File(filename, 'r')['y_data_noisefree'][idx]),(self.batch_size,self.n_channels,self.y_dim)),[0,2,1])
        rand_pars = h5py.File(filename, 'r')['rand_pars'][:]

        # convert ra to hour angle
        x = convert_sky(x, params['ref_geocent_time'], rand_pars)

        # Normalise the source parameters
        for i,k in enumerate(rand_pars):
            par_min = k.decode('utf-8') + '_min'
            par_max = k.decode('utf-8') + '_max'

            if par_min=='psi_min':
                x[:,i] = np.remainder(x[:,i], np.pi)

            # normalize by bounds
            x[:,i]=(x[:,i] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])

        # extract inference parameters from all source parameters loaded earlier
        idx = []
        infparlist = ''
        for k in params['inf_pars']:
            infparlist = infparlist + k + ', '
            for i,q in enumerate(rand_pars):
                m = q.decode('utf-8')
                if k==m:
                    idx.append(i)
        x = tf.cast(x[:,idx],dtype=tf.float32)
        y = tf.cast(y,dtype=tf.float32)

        # randomise distance
        old_d = bounds['luminosity_distance_min'] + tf.boolean_mask(x,dist_mask,axis=1)*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
        new_x = tf.random.uniform(shape=tf.shape(old_d), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
        new_d = bounds['luminosity_distance_min'] + new_x*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
        x = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_dist_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_dist_mask),axis=1)
        dist_scale = tf.tile(tf.expand_dims(old_d/new_d,axis=1),(1,tf.shape(y)[1],1))

        # augment the data for this chunk - randomize phase and arrival time
        old_phase = bounds['phase_min'] + tf.boolean_mask(x,phase_mask,axis=1)*(bounds['phase_max'] - bounds['phase_min'])
        new_x = tf.random.uniform(shape=tf.shape(old_phase), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
        new_phase = bounds['phase_min'] + new_x*(bounds['phase_max'] - bounds['phase_min'])
        x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_phase_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_phase_mask),axis=1)
        old_geocent = bounds['geocent_time_min'] + tf.boolean_mask(x,geocent_mask,axis=1)*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
        new_x = tf.random.uniform(shape=tf.shape(old_geocent), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
        new_geocent = bounds['geocent_time_min'] + new_x*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
        x = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_geocent_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_geocent_mask),axis=1)
        phase_correction = -1.0*tf.complex(tf.cos(new_phase-old_phase),tf.sin(new_phase-old_phase))
        phase_correction = tf.tile(tf.expand_dims(phase_correction,axis=1),(1,3,tf.shape(y)[1]/2 + 1))
        time_correction = -2.0*np.pi*self.fvec*(new_geocent-old_geocent)
        time_correction = tf.complex(tf.cos(time_correction),tf.sin(time_correction))
        time_correction = tf.tile(tf.expand_dims(time_correction,axis=1),(1,3,1))
        y_fft = tf.signal.rfft(tf.transpose(y,[0,2,1]))*phase_correction*time_correction
        y = tf.transpose(tf.signal.irfft(y_fft),[0,2,1])

        return x, y

def get_param_index(all_pars,pars,sky_extra=None):
    """ 
    Get the list index of requested source parameter types
    """
    # identify the indices of wrapped and non-wrapped parameters - clunky code
    mask = []
    idx = []

    # loop over inference params
    for i,p in enumerate(all_pars):

        # loop over wrapped params
        flag = False
        for q in pars:
            if p==q:
                flag = True    # if inf params is a wrapped param set flag

        # record the true/false value for this inference param
        if flag==True:
            mask.append(True)
            idx.append(i)
        elif flag==False:
            mask.append(False)

    if sky_extra is not None:
        if sky_extra:
            mask.append(True)
            idx.append(len(all_pars))
        else:
            mask.append(False)

    return mask, idx, np.sum(mask)

def convert_sky(data, geocent_time, pars, single=False):
    """
    Converts right ascension to hour angle or the reverse
    """
    greenwich = coord.EarthLocation.of_site('greenwich')
    t = Time(geocent_time, format='gps', location=greenwich)
    t = t.sidereal_time('mean', 'greenwich').radian

    # compute single instance
    if single:
        return np.remainder(t - data,2.0*np.pi)

    for i,k in enumerate(pars):
        try:
            k = k.decode('utf-8')
        except:
            pass
        if k == 'ra':
            ra_idx = i

    # Check if RA exist
    try:
        ra_idx
    except NameError:
        print('...... RA is fixed. Not converting RA to hour angle or the reverse.')
    else:
        # Iterate over all training samples and convert to hour angle
        for i in range(data.shape[0]):
            data[i,ra_idx] = np.remainder(t - data[i,ra_idx],2.0*np.pi)

    return data

def load_data(params,bounds,fixed_vals,input_dir,inf_pars,test_data=False,max_samples=np.inf,silent=False):
    """ Function to load either training or testing data.
    """

    # Get list of all training/testing files and define dictionary to store values in files
    if type("%s" % input_dir) is str:
        dataLocations = ["%s" % input_dir]
    else:
        print('ERROR: input directory not a string')
        exit(0)

    # Sort files
    filenames = sorted(os.listdir(dataLocations[0]))
    if test_data:   # if doing test data then make sure the ordering is correct
        filenames = sorted(os.listdir(dataLocations[0]),key=lambda x: int(x[:-5].split('_')[1]))

    # If loading by chunks, randomly shuffle list of training/testing filenames
    if params['load_by_chunks'] == True and not test_data:
        nfiles = np.min([int(params['load_chunk_size']/float(params['tset_split'])),len(filenames)])
        files_idx = np.random.randint(0,len(filenames),nfiles) 
        filenames= np.array(filenames)[files_idx]
        if not silent:
            print('...... shuffled filenames since we are loading in by chunks')

    # Iterate over all training/testing files and store source parameters, time series and SNR info in dictionary
    data={'x_data': [], 'y_data_noisefree': [], 'y_data_noisy': [], 'rand_pars': [], 'snrs': []}
    for filename in filenames:
        try:
            data['x_data'].append(h5py.File(dataLocations[0]+'/'+filename, 'r')['x_data'][:])
            data['y_data_noisefree'].append(h5py.File(dataLocations[0]+'/'+filename, 'r')['y_data_noisefree'][:])
            if test_data:
                data['y_data_noisy'].append(h5py.File(dataLocations[0]+'/'+filename, 'r')['y_data_noisy'][:])
            data['rand_pars'] = h5py.File(dataLocations[0]+'/'+filename, 'r')['rand_pars'][:]
            data['snrs'].append(h5py.File(dataLocations[0]+'/'+filename, 'r')['snrs'][:])
            if not silent:
                print('...... Loaded file ' + dataLocations[0] + '/' + filename)
        except OSError:
            print('Could not load requested file')
            continue
    ndet = np.array(data['y_data_noisefree']).ndim
    if ndet == 3:
        data['y_data_noisefree'] = np.expand_dims(np.array(data['y_data_noisefree']),axis=0)
        data['y_data_noisy'] = np.expand_dims(np.array(data['y_data_noisy']),axis=0)
    data['x_data'] = np.concatenate(np.array(data['x_data']), axis=0).squeeze()
    data['y_data_noisefree'] = np.transpose(np.concatenate(np.array(data['y_data_noisefree']), axis=0),[0,2,1])
    data['snrs'] = np.concatenate(np.array(data['snrs']), axis=0).reshape(-1,ndet)
    if test_data:  # trim them to desired length
        data['x_data'] = data['x_data'][:max_samples]
        data['snrs'] = data['snrs'][:max_samples]
        data['y_data_noisefree'] = data['y_data_noisefree'][:max_samples]
        data['y_data_noisy'] = np.transpose(np.concatenate(np.array(data['y_data_noisy']), axis=0),[0,2,1])[:max_samples]

    # convert ra to hour angle
    data['x_data'] = convert_sky(data['x_data'], params['ref_geocent_time'], data['rand_pars'])

    # Normalise the source parameters
    for i,k in enumerate(data['rand_pars']):
        par_min = k.decode('utf-8') + '_min'
        par_max = k.decode('utf-8') + '_max'
        
        if par_min=='psi_min':
            data['x_data'][:,i] = np.remainder(data['x_data'][:,i], np.pi)
        
        # normalize by bounds
        data['x_data'][:,i]=(data['x_data'][:,i] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])

    # extract inference parameters from all source parameters loaded earlier
    idx = []
    infparlist = ''
    for k in inf_pars:
        infparlist = infparlist + k + ', '
        for i,q in enumerate(data['rand_pars']):
            m = q.decode('utf-8')
            if k==m:
                idx.append(i)
    data['x_data'] = tf.cast(data['x_data'][:,idx],dtype=tf.float32)
    if not silent:
        print('...... {} will be inferred'.format(infparlist))

    return data['x_data'], tf.cast(data['y_data_noisefree'],dtype=tf.float32), tf.cast(data['y_data_noisy'],dtype=tf.float32), data['snrs']

def load_samples(params,sampler):
    """
    read in pre-computed posterior samples
    """
    if type("%s" % params['pe_dir']) is str:
        # load generated samples back in
        dataLocations = '%s_%s' % (params['pe_dir'],sampler)
        print('... looking in {} for posterior samples'.format(dataLocations))
    else:
        print('ERROR: input samples directory not a string')
        exit(0)

    # Iterate over requested number of testing samples to use
    for i in range(params['r']):

        filename = '%s/%s_%d.h5py' % (dataLocations,params['bilby_results_label'],i)
        if not os.path.isfile(filename):
            print('... unable to find file {}. Exiting.'.format(filename))
            exit(0)

        print('... Loading test sample -> ' + filename)
        data_temp = {}
        n = 0

        # Retrieve all source parameters to do inference on
        for q in params['bilby_pars']:
            p = q + '_post'
            par_min = q + '_min'
            par_max = q + '_max'
            data_temp[p] = h5py.File(filename, 'r')[p][:]
            if p == 'psi_post':
                print('dealing with psi in samples')
                data_temp[p] = np.remainder(data_temp[p],np.pi)
            if p == 'geocent_time_post':
                data_temp[p] = data_temp[p] - params['ref_geocent_time']
            data_temp[p] = (data_temp[p] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])
            Nsamp = data_temp[p].shape[0]
            n = n + 1
        print('... read in {} samples from {}'.format(Nsamp,filename))

        # place retrieved source parameters in numpy array rather than dictionary
        j = 0
        XS = np.zeros((Nsamp,n))
        for p,d in data_temp.items():
            XS[:,j] = d
            j += 1
        print('... put the samples in an array')

        # Append test sample posteriors to existing array of other test sample posteriors
        rand_idx_posterior = np.random.randint(0,Nsamp,params['n_samples'])
        if i == 0:
            XS_all = np.expand_dims(XS[rand_idx_posterior,:], axis=0)
        else:
            XS_all = np.vstack((XS_all,np.expand_dims(XS[rand_idx_posterior,:], axis=0)))
        print('... appended {} samples to the total'.format(params['n_samples']))
    return XS_all

def plot_losses(train_loss, val_loss, epoch, run='testing'):
    """
    plots the losses
    """
    plt.figure(figsize=(14,14))
    plt.semilogx(np.arange(1,epoch+1),train_loss[:epoch,0],'b',label='RECON')
    plt.semilogx(np.arange(1,epoch+1),train_loss[:epoch,1],'r',label='KL')
    plt.semilogx(np.arange(1,epoch+1),train_loss[:epoch,2],'g',label='TOTAL')
    plt.semilogx(np.arange(1,epoch+1),val_loss[:epoch,0],'--b',alpha=0.5)
    plt.semilogx(np.arange(1,epoch+1),val_loss[:epoch,1],'--r',alpha=0.5)
    plt.semilogx(np.arange(1,epoch+1),val_loss[:epoch,2],'--g',alpha=0.5)
    plt.xlabel('epoch')
    plt.ylabel('loss')
    plt.legend()
    plt.grid()
    plt.ylim([np.min(1.1*train_loss[int(0.1*epoch):epoch,:]),np.max(1.1*train_loss[int(0.1*epoch):epoch,:])])
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/loss.png' % (run))
    if epoch>1:
        plt.xlim([max(1,epoch-100),epoch+1])
        plt.ylim([np.min(train_loss[max(1,epoch-100):epoch,2])-1.0,np.max(train_loss[max(1,epoch-100):epoch,2])+1.0])
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/loss_latest_total.png' % (run))
    plt.close()

    # save loss data to text file
    loss_file = '/data/www.astro/chrism/vitamin_c/{}/loss.txt'.format(run)
    f = open(loss_file,'a+')
    data = tf.reshape(np.concatenate([train_loss[epoch-1,:],val_loss[epoch-1,:]],axis=0),(1,6))
    f.write('%d %f %f %f %f %f %f\n' % (epoch,data[0,0],data[0,1],data[0,2],data[0,3],data[0,4],data[0,5]))
    f.close()

 
def plot_KL(KL_samples, step, run='testing'):
    """
    plots the KL evolution
    """
    # arrives in shape n_kl,n_test,3
    N = KL_samples.shape[0]
    KL_samples = np.transpose(KL_samples,[2,1,0])   # re-order axes
    ls = ['-','--',':']
    c = ['C0','C1','C2','C3']
    fig, axs = plt.subplots(3, sharex=True, sharey=True, figsize=(6.4,14.4))
    for i,kl_s in enumerate(KL_samples):   # loop over samplers
        for j,kl in enumerate(kl_s):     # loop over test cases
            axs[i].semilogx(np.arange(1,N+1)*step,kl,ls[i],color=c[j])
            axs[i].plot(N*step,kl[-1],'.',color=c[j])
            axs[i].grid()
    plt.xlabel('epoch')
    plt.ylabel('KL')
    plt.ylim([-0.2,1.0])
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/kl.png' % (run))
    plt.close()

def plot_mode_posterior(samples,x_truth,snrs,epoch,idx,run='testing'):
    """
    plots the posteriors in individual latent space modes
    """

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=None,
                    levels=[0.90], density=True,
                    plot_density=False, plot_datapoints=False,
                    max_n_ticks=3)

    col = ['tab:blue',
           'tab:orange',
           'tab:green',
           'tab:red',
           'tab:purple',
           'tab:brown',
           'tab:pink',
           'tab:gray',
           'tab:olive',
           'tab:cyan']
 
    for i, sample in enumerate(samples):
        true_x = np.zeros(inf_ol_len)
        true_XS = np.zeros([sample.shape[0],inf_ol_len])
        ol_pars = []
        #par_range = np.zeros([inf_ol_len,2])
        cnt = 0
        for inf_idx in inf_ol_idx:
            inf_par = params['inf_pars'][inf_idx]
            true_XS[:,cnt] = (sample[:,inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par+'_min']
            true_x[cnt] = (x_truth[inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par + '_min']
            ol_pars.append(inf_par)
            #par_range[cnt,0] = bounds[inf_par+'_min']
            #par_range[cnt,1] = bounds[inf_par+'_max']
            cnt += 1
        parnames = []
        for k_idx,k in enumerate(params['rand_pars']):
            if np.isin(k, ol_pars):
                parnames.append(params['corner_labels'][k])

        # convert to RA
        true_XS = convert_sky(true_XS,params['ref_geocent_time'],ol_pars)
        true_x = convert_sky(np.reshape(true_x,[1,true_XS.shape[1]]),params['ref_geocent_time'],ol_pars).flatten()

        hist_kwargs = dict(density=True,color=col[np.remainder(i,len(col))])
        if i==0:
            figure = corner.corner(true_XS, **defaults_kwargs,labels=parnames,
                           color=col[np.remainder(i,len(col))], truths=true_x,
                           show_titles=True, hist_kwargs=hist_kwargs)
        else:
            corner.corner(true_XS,**defaults_kwargs,
                           color=col[np.remainder(i,len(col))],
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs)

    snr_tot = np.sqrt(np.sum(snrs**2))
    plt.annotate('SNR = {:.3f} [{:.3f},{:.3f},{:.3f}]'.format(snr_tot,snrs[0],snrs[1],snrs[2]),(0.6,0.95),xycoords='figure fraction',fontsize=18)
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/modes_posterior_epoch_%d_event_%d.png' % (run,epoch,idx))
    plt.close()

def plot_posterior(samples,x_truth,snrs,epoch,idx,run='testing',all_other_samples=None):
    """
    plots the posteriors
    """

    # trim samples from outside the cube
    mask = []
    for s in samples:
        if (np.all(s>=0.0) and np.all(s<=1.0)):
            mask.append(True)
        else:
            mask.append(False)
    samples = tf.boolean_mask(samples,mask,axis=0)
    print('identified {} good samples'.format(samples.shape[0]))
    print(np.array(all_other_samples).shape)
    if samples.shape[0]<100:
        return [-1.0,-1.0,-1.0]

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=[0.16, 0.84],
                    levels=(0.68,0.90,0.95), density=True,
                    plot_density=False, plot_datapoints=True,
                    max_n_ticks=3)

    # 1-d hist kwargs for normalisation
    hist_kwargs = dict(density=True,color='tab:red')
    hist_kwargs_other = dict(density=True,color='tab:blue')
    hist_kwargs_other2 = dict(density=True,color='tab:green')

    if all_other_samples is not None:
        KL_est = []
        for i, other_samples in enumerate(all_other_samples):
            true_post = np.zeros([other_samples.shape[0],bilby_ol_len])
            true_x = np.zeros(inf_ol_len)
            true_XS = np.zeros([samples.shape[0],inf_ol_len])
            ol_pars = []
            cnt = 0
            for inf_idx,bilby_idx in zip(inf_ol_idx,bilby_ol_idx):
                inf_par = params['inf_pars'][inf_idx]
                bilby_par = params['bilby_pars'][bilby_idx]
                true_XS[:,cnt] = (samples[:,inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par+'_min']
                true_post[:,cnt] = (other_samples[:,bilby_idx] * (bounds[bilby_par+'_max'] - bounds[bilby_par+'_min'])) + bounds[bilby_par + '_min']
                true_x[cnt] = (x_truth[inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par + '_min']
                ol_pars.append(inf_par)
                cnt += 1
            parnames = []
            for k_idx,k in enumerate(params['rand_pars']):
                if np.isin(k, ol_pars):
                    parnames.append(params['corner_labels'][k])

            # convert to RA
            true_XS = convert_sky(true_XS,params['ref_geocent_time'],ol_pars)
            true_x = convert_sky(np.reshape(true_x,[1,true_XS.shape[1]]),params['ref_geocent_time'],ol_pars).flatten()
            old_true_post = true_post                 

            samples_file = '/data/www.astro/chrism/vitamin_c/{}/posterior_samples_epoch_{}_event_{}_vit.txt'.format(run,epoch,idx)
            np.savetxt(samples_file,true_XS)

            # compute KL estimate
            idx1 = np.random.randint(0,true_XS.shape[0],2000)
            idx2 = np.random.randint(0,true_post.shape[0],2000)
            try:
                current_KL = -1.0 #0.5*(estimate(true_XS[idx1,:],true_post[idx2,:],n_jobs=4) + estimate(true_post[idx2,:],true_XS[idx1,:],n_jobs=4))
            except:
                current_KL = -1.0
                pass
            KL_est.append(current_KL)

            other_samples_file = '/data/www.astro/chrism/vitamin_c/{}/posterior_samples_epoch_{}_event_{}_{}.txt'.format(run,epoch,idx,i)
            np.savetxt(other_samples_file,true_post)

            if i==0:
                figure = corner.corner(true_post, **defaults_kwargs,labels=parnames,
                           color='tab:blue',
                           show_titles=True, hist_kwargs=hist_kwargs_other)
            else:

                # compute KL estimate
                idx1 = np.random.randint(0,old_true_post.shape[0],2000)
                idx2 = np.random.randint(0,true_post.shape[0],2000)
                try:
                    current_KL = -1.0 #0.5*(estimate(old_true_post[idx1,:],true_post[idx2,:],n_jobs=4) + estimate(true_post[idx2,:],old_true_post[idx1,:],n_jobs=4))
                except:
                    current_KL = -1.0
                    pass
                KL_est.append(current_KL)

                corner.corner(true_post,**defaults_kwargs,
                           color='tab:green',
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs_other2)
        
        for j,KL in enumerate(KL_est):    
            plt.annotate('KL = {:.3f}'.format(KL),(0.2,0.95-j*0.02),xycoords='figure fraction',fontsize=18)

        snr_tot = np.sqrt(np.sum(snrs**2))
        plt.annotate('SNR = {:.3f} [{:.3f},{:.3f},{:.3f}]'.format(snr_tot,snrs[0],snrs[1],snrs[2]),(0.6,0.95),xycoords='figure fraction',fontsize=18)
        corner.corner(true_XS,**defaults_kwargs,
                           color='tab:red',
                           fill_contours=False, truths=true_x,
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs)
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/comp_posterior_epoch_%d_event_%d.png' % (run,epoch,idx))
        plt.close()
        return KL_est

    else:
        # Get corner parnames to use in plotting labels
        parnames = []
        for k_idx,k in enumerate(params['rand_pars']):
            if np.isin(k, params['inf_pars']):
                parnames.append(params['corner_labels'][k])
        # un-normalise full inference parameters
        full_true_x = np.zeros(len(params['inf_pars']))
        new_samples = np.zeros([samples.shape[0],len(params['inf_pars'])])
        for inf_par_idx,inf_par in enumerate(params['inf_pars']):
            new_samples[:,inf_par_idx] = (samples[:,inf_par_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par+'_min']
            full_true_x[inf_par_idx] = (x_truth[inf_par_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par + '_min']
        new_samples = convert_sky(new_samples,params['ref_geocent_time'],params['inf_pars'])
        full_true_x = convert_sky(np.reshape(full_true_x,[1,samples.shape[1]]),params['ref_geocent_time'],params['inf_pars']).flatten()       

        figure = corner.corner(new_samples,**defaults_kwargs,labels=parnames,
                           color='tab:red',
                           fill_contours=False, truths=full_true_x,
                           show_titles=True, hist_kwargs=hist_kwargs)
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/full_posterior_epoch_%d_event_%d.png' % (run,epoch,idx))
        plt.close()
    return -1.0

def plot_latent(mu_r1, logw_r1, z_r1, mu_q, z_q, epoch, idx, run='testing'):

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=[0.16, 0.84],
                    levels=(0.68,0.90,0.95), density=True,
                    plot_density=False, plot_datapoints=True,
                    max_n_ticks=3)

    # 1-d hist kwargs for normalisation
    hist_kwargs = dict(density=True,color='tab:red')
    hist_kwargs_other = dict(density=True,color='tab:blue')

    figure = corner.corner(z_q, **defaults_kwargs,
                           color='tab:blue',
                           show_titles=True, hist_kwargs=hist_kwargs_other)
    corner.corner(z_r1,**defaults_kwargs,
                           color='tab:red',
                           fill_contours=True,
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs)
    # Extract the axes
    z_dim = z_r1.shape[1]
    axes = np.array(figure.axes).reshape((z_dim, z_dim))

    # Loop over the histograms
    for yi in range(z_dim):
        for xi in range(yi):
            ax = axes[yi, xi]
            ax.plot(mu_r1[0,:,xi], mu_r1[0,:,yi], "sr")
            ax.plot(mu_q[0,xi], mu_q[0,yi], "sb")
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/latent_epoch_%d_event_%d.png' % (run,epoch,idx))
    plt.close()

    # plot the weights
    plt.figure()
    plt.step(np.arange(params['n_modes']), logw_r1[0,:], where='mid')
    plt.xlabel('z dimension')
    plt.ylabel('log(w)')
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/latent_logweight_epoch_%d_event_%d.png' % (run,epoch,idx))
    plt.close()
    plt.figure()
    w = np.exp(logw_r1[0,:] - np.max(logw_r1[0,:]))
    w = w/np.sum(w)
    plt.step(np.arange(params['n_modes']), w, where='mid')
    plt.xlabel('z dimension')
    plt.ylabel('w')
    plt.ylim([0,np.max(w)*1.1])
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/latent_weight_epoch_%d_event_%d.png' % (run,epoch,idx))
    plt.close()

params = '/home/chrism/vitamin_c/params.json'
bounds = '/home/chrism/vitamin_c//bounds.json'
fixed_vals = '/home/chrism/vitamin_c/fixed_vals.json'
#run = time.ctime().replace(' ', '-')
run = time.strftime('%y-%m-%d-%X-%Z')
EPS = 1e-6

path = os.path.join('/data/www.astro/chrism/vitamin_c', run) 
os.mkdir(path) 
shutil.copy('./vitamin_c.py',path)
shutil.copy('./params.json',path)

# Load parameters files
with open(params, 'r') as fp:
    params = json.load(fp)
with open(bounds, 'r') as fp:
    bounds = json.load(fp)
with open(fixed_vals, 'r') as fp:
    fixed_vals = json.load(fp)

# if doing hour angle, use hour angle bounds on RA
#bounds['ra_min'] = convert_sky(bounds['ra_min'],params['ref_geocent_time'],None,single=True)
#bounds['ra_max'] = convert_sky(bounds['ra_max'],params['ref_geocent_time'],None,single=True)
#print('... converted RA bounds to hour angle')
inf_ol_mask, inf_ol_idx, inf_ol_len = get_param_index(params['inf_pars'],params['bilby_pars'])
bilby_ol_mask, bilby_ol_idx, bilby_ol_len = get_param_index(params['bilby_pars'],params['inf_pars'])

# identify the indices of different sets of physical parameters
vonmise_mask, vonmise_idx_mask, vonmise_len = get_param_index(params['inf_pars'],params['vonmise_pars'])
gauss_mask, gauss_idx_mask, gauss_len = get_param_index(params['inf_pars'],params['gauss_pars'])
sky_mask, sky_idx_mask, sky_len = get_param_index(params['inf_pars'],params['sky_pars'])
ra_mask, ra_idx_mask, ra_len = get_param_index(params['inf_pars'],['ra'])
dec_mask, dec_idx_mask, dec_len = get_param_index(params['inf_pars'],['dec'])
m1_mask, m1_idx_mask, m1_len = get_param_index(params['inf_pars'],['mass_1'])
m2_mask, m2_idx_mask, m2_len = get_param_index(params['inf_pars'],['mass_2'])
#idx_mask = np.argsort(gauss_idx_mask + vonmise_idx_mask + m1_idx_mask + m2_idx_mask + sky_idx_mask) # + dist_idx_mask)
idx_mask = np.argsort(m1_idx_mask + m2_idx_mask + gauss_idx_mask + vonmise_idx_mask) # + sky_idx_mask)
dist_mask, dist_idx_mask, dist_len = get_param_index(params['inf_pars'],['luminosity_distance'])
not_dist_mask, not_dist_idx_mask, not_dist_len = get_param_index(params['inf_pars'],['mass_1','mass_2','psi','phase','geocent_time','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
phase_mask, phase_idx_mask, phase_len = get_param_index(params['inf_pars'],['phase'])
not_phase_mask, not_phase_idx_mask, not_phase_len = get_param_index(params['inf_pars'],['mass_1','mass_2','luminosity_distance','psi','geocent_time','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
geocent_mask, geocent_idx_mask, geocent_len = get_param_index(params['inf_pars'],['geocent_time'])
not_geocent_mask, not_geocent_idx_mask, not_geocent_len = get_param_index(params['inf_pars'],['mass_1','mass_2','luminosity_distance','psi','phase','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
xyz_mask, xyz_idx_mask, xyz_len = get_param_index(params['inf_pars'],['luminosity_distance','ra','dec'])
not_xyz_mask, not_xyz_idx_mask, not_xyz_len = get_param_index(params['inf_pars'],['mass_1','mass_2','psi','phase','geocent_time','theta_jn','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
idx_xyz_mask = np.argsort(xyz_idx_mask + not_xyz_idx_mask)
idx_dist_mask = np.argsort(not_dist_idx_mask + dist_idx_mask)
idx_phase_mask = np.argsort(not_phase_idx_mask + phase_idx_mask)
idx_geocent_mask = np.argsort(not_geocent_idx_mask + geocent_idx_mask)
print(xyz_mask)
print(not_xyz_mask)
print(idx_xyz_mask)
#masses_len = m1_len + m2_len
print(params['inf_pars'])
print(vonmise_mask,vonmise_idx_mask)
print(gauss_mask,gauss_idx_mask)
print(m1_mask,m1_idx_mask)
print(m2_mask,m2_idx_mask)
print(sky_mask,sky_idx_mask)
print(idx_mask)

# define which gpu to use during training
gpu_num = str(params['gpu_num'])   
os.environ["CUDA_VISIBLE_DEVICES"]=gpu_num
print('... running on GPU {}'.format(gpu_num))

# Let GPU consumption grow as needed
config = tf.compat.v1.ConfigProto()
config.gpu_options.allow_growth = True
session = tf.compat.v1.Session(config=config)
print('... letting GPU consumption grow as needed')

# load the data
#training_generator = DataGenerator(params['train_set_dir'], samples_per_file=1000, batch_size=512, x_dim=15, y_dim=1024, n_channels=3, shuffle=True)
#x_data_train,y_data_train = training_generator.__getitem__(0)

x_data_train, y_data_train, _, snrs_train = load_data(params,bounds,fixed_vals,params['train_set_dir'],params['inf_pars'])

fig, ax = plt.subplots(3,3,sharex=True,sharey=True,figsize=(15,15))
idx = 0
t = params['duration']*np.arange(params['ndata'])/float(params['ndata'])
for i in range(3):
    for j in range(3):
        ax[i,j].plot(t,y_data_train[idx,:,0],'b')
        ax[i,j].plot(t,y_data_train[idx,:,1],'r')
        ax[i,j].plot(t,y_data_train[idx,:,2],'g')
        idx += 1
plt.savefig('/data/www.astro/chrism/vitamin_c/{}/train_signals.png'.format(run))
plt.close()

# augment the data for this chunk - randomize phase and arrival time
#old_phase = bounds['phase_min'] + tf.boolean_mask(x_data_train,phase_mask,axis=1)*(bounds['phase_max'] - bounds['phase_min'])
#new_x = tf.random.uniform(shape=tf.shape(old_phase), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
#new_phase = bounds['phase_min'] + new_x*(bounds['phase_max'] - bounds['phase_min'])
#x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_phase_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_phase_mask),axis=1)
#old_geocent = bounds['geocent_time_min'] + tf.boolean_mask(x_data_train,geocent_mask,axis=1)*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
#new_x = tf.random.uniform(shape=tf.shape(old_geocent), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
#new_geocent = bounds['geocent_time_min'] + new_x*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
#x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_geocent_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_geocent_mask),axis=1)
#phase_correction = -1.0*tf.complex(tf.cos(new_phase-old_phase),tf.sin(new_phase-old_phase))
#phase_correction = tf.tile(tf.expand_dims(phase_correction,axis=1),(1,3,tf.shape(y_data_train)[1]/2 + 1))
#fvec = tf.range(params['ndata']/2 + 1)/params['duration']
#time_correction = -2.0*np.pi*fvec*(new_geocent-old_geocent)
#time_correction = tf.complex(tf.cos(time_correction),tf.sin(time_correction))
#time_correction = tf.tile(tf.expand_dims(time_correction,axis=1),(1,3,1))
#y_data_train_fft = tf.signal.rfft(tf.transpose(y_data_train,[0,2,1]))*phase_correction*time_correction
#y_data_train = tf.transpose(tf.signal.irfft(y_data_train_fft),[0,2,1])

#print(x_data_train[1,:])
#print(old_geocent[1],new_geocent[1])
#plt.plot(y_data_train[1,:,0],'r')
#plt.savefig('/data/www.astro/chrism/vitamin_c/time_phase_shift.png')
#plt.close()
#exit(0)

x_data_val, y_data_val, _, snrs_val = load_data(params,bounds,fixed_vals,params['val_set_dir'],params['inf_pars'])

fig, ax = plt.subplots(3,3,sharex=True,sharey=True,figsize=(15,15))
idx = 0
t = params['duration']*np.arange(params['ndata'])/float(params['ndata'])
for i in range(3):
    for j in range(3):
        ax[i,j].plot(t,y_data_val[idx,:,0],'b')
        ax[i,j].plot(t,y_data_val[idx,:,1],'r')
        ax[i,j].plot(t,y_data_val[idx,:,2],'g')
        idx += 1
plt.savefig('/data/www.astro/chrism/vitamin_c/{}/val_signals.png'.format(run))
plt.close()

x_data_test, y_data_test_noisefree, y_data_test, snrs_test = load_data(params,bounds,fixed_vals,params['test_set_dir'],params['inf_pars'],max_samples=params['r'],test_data=True)

fig, ax = plt.subplots(2,2,sharex=True,sharey=True,figsize=(15,15))
idx = 0
t = params['duration']*np.arange(params['ndata'])/float(params['ndata'])
for i in range(2):
    for j in range(2):
        ax[i,j].plot(t,y_data_test_noisefree[idx,:,0],'b')
        ax[i,j].plot(t,y_data_test_noisefree[idx,:,1],'r')
        ax[i,j].plot(t,y_data_test_noisefree[idx,:,2],'g')
        idx += 1
plt.savefig('/data/www.astro/chrism/vitamin_c/{}/test_signals.png'.format(run))
plt.close()

fig, ax = plt.subplots(2,2,sharex=True,sharey=True,figsize=(15,15))
idx = 0
t = params['duration']*np.arange(params['ndata'])/float(params['ndata'])
for i in range(2):
    for j in range(2):
        ax[i,j].plot(t,y_data_test[idx,:,0],'b')
        ax[i,j].plot(t,y_data_test[idx,:,1],'r')
        ax[i,j].plot(t,y_data_test[idx,:,2],'g')
        idx += 1
plt.savefig('/data/www.astro/chrism/vitamin_c/{}/test_data.png'.format(run))
plt.close()

x_data_pp, y_data_pp_noisefree, y_data_pp, snrs_pp = load_data(params,bounds,fixed_vals,params['pp_set_dir'],params['inf_pars'],max_samples=100,test_data=True)

fig, ax = plt.subplots(3,3,sharex=True,sharey=True,figsize=(15,15))
idx = 0
t = params['duration']*np.arange(params['ndata'])/float(params['ndata'])
for i in range(3):
    for j in range(3):
        ax[i,j].plot(t,y_data_pp_noisefree[idx,:,0],'b')
        ax[i,j].plot(t,y_data_pp_noisefree[idx,:,1],'r')
        ax[i,j].plot(t,y_data_pp_noisefree[idx,:,2],'g')
        idx += 1
plt.savefig('/data/www.astro/chrism/vitamin_c/{}/pp_signals.png'.format(run))
plt.close()

fig, ax = plt.subplots(3,3,sharex=True,sharey=True,figsize=(15,15))
idx = 0
t = params['duration']*np.arange(params['ndata'])/float(params['ndata'])
for i in range(3):
    for j in range(3):
        ax[i,j].plot(t,y_data_pp[idx,:,0],'b')
        ax[i,j].plot(t,y_data_pp[idx,:,1],'r')
        ax[i,j].plot(t,y_data_pp[idx,:,2],'g')
        idx += 1
plt.savefig('/data/www.astro/chrism/vitamin_c/{}/pp_data.png'.format(run))
plt.close()

fig,ax = plt.subplots(14,14,figsize=(15,15))
for i in range(14):
    for j in range(14):
        if j<i:
            ax[i,j].plot(x_data_train[:,j],x_data_train[:,i],',b')
            ax[i,j].plot(x_data_val[:,j],x_data_val[:,i],',g')
            ax[i,j].plot(x_data_pp[:,j],x_data_pp[:,i],',r')
            ax[i,j].plot(x_data_test[:,j],x_data_test[:,i],'xk')
            ax[i,j].set_xlabel(params['inf_pars'][j])
            ax[i,j].set_ylabel(params['inf_pars'][i])
plt.savefig('/data/www.astro/chrism/vitamin_c/{}/train_prior.png'.format(run))
plt.close()

print(x_data_val.shape)

snrs_test = np.reshape(snrs_test,[-1,3])

plt.figure()
#plt.hist(np.sqrt(np.sum(snrs_train**2,axis=1)),100,density=True,histtype='stepfilled',alpha=0.5)
plt.hist(np.sqrt(np.sum(snrs_val**2,axis=1)),100,density=True,histtype='stepfilled',alpha=0.5)
plt.plot(np.sqrt(np.sum(snrs_pp**2,axis=1)),np.zeros(snrs_pp.shape[0]),'ok')
plt.plot(np.sqrt(np.sum(snrs_test**2,axis=1)),np.zeros(snrs_test.shape[0]),'sr')
plt.plot(np.sqrt(np.sum(snrs_pp[:4,:]**2,axis=1)),np.zeros(4),'ob')
plt.savefig('/data/www.astro/chrism/vitamin_c/%s/snrs.png' % (run))
plt.close()

# load precomputed samples
bilby_samples = np.array([load_samples(params,'dynesty1'),load_samples(params,'cpnest1')])
print('bilby samples shape = '.format(bilby_samples.shape))

train_size = params['load_chunk_size']
batch_size = 512
val_size = batch_size*19
test_size = 8
plot_cadence = 1000

train_dataset = (tf.data.Dataset.from_tensor_slices((x_data_train,y_data_train))
                 .shuffle(train_size).batch(batch_size))
val_dataset = (tf.data.Dataset.from_tensor_slices((x_data_val,y_data_val))
                .take(val_size).batch(batch_size))
print(x_data_test.shape,y_data_test.shape,snrs_test.shape)
test_dataset = (tf.data.Dataset.from_tensor_slices((x_data_test,y_data_test,snrs_test))
                .batch(1))
print(x_data_pp.shape,y_data_pp.shape,snrs_pp.shape)
pp_dataset = (tf.data.Dataset.from_tensor_slices((x_data_pp,y_data_pp,snrs_pp))
                .batch(1))

class CVAE(tf.keras.Model):
    """Convolutional variational autoencoder."""

    def __init__(self, x_dim, y_dim, n_channels, z_dim, n_modes):
        super(CVAE, self).__init__()
        self.z_dim = z_dim
        self.n_modes = n_modes
        self.x_modes = 1   # hardcoded for testing
        self.x_dim = x_dim
        self.y_dim = y_dim
        self.n_channels = n_channels
        self.act = 'relu'

        # the r1 network
        all_input_y = tf.keras.Input(shape=(self.y_dim, self.n_channels))
        r1c = tf.keras.layers.Conv1D(filters=128, kernel_size=64, strides=1, activation=self.act)(all_input_y)
        r1c = tf.keras.layers.Conv1D(filters=128, kernel_size=64, strides=4, activation=self.act)(r1c)
        r1c = tf.keras.layers.Conv1D(filters=128, kernel_size=32, strides=1, activation=self.act)(r1c)
        r1c = tf.keras.layers.Conv1D(filters=96, kernel_size=32, strides=2, activation=self.act)(r1c)
        r1c = tf.keras.layers.Conv1D(filters=96, kernel_size=16, strides=1, activation=self.act)(r1c)
        r1c = tf.keras.layers.Conv1D(filters=96, kernel_size=16, strides=2, activation=self.act)(r1c)
        r1c = tf.keras.layers.Flatten()(r1c)
        r1 = tf.keras.layers.Dense(2048,activation=self.act)(r1c)
        r1 = tf.keras.layers.Dense(2048,activation=self.act)(r1)
        r1 = tf.keras.layers.Dense(1024,activation=self.act)(r1)
        r1mu = tf.keras.layers.Dense(self.z_dim*self.n_modes,use_bias=True,bias_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=1.0))(r1)
        r1logvar = tf.keras.layers.Dense(self.z_dim*self.n_modes,use_bias=True,bias_initializer=tf.keras.initializers.RandomUniform(minval=-4.0, maxval=0.0))(r1)
        r1w = tf.keras.layers.Dense(self.n_modes,use_bias=True,bias_initializer='Zeros')(r1)
        r1 = tf.keras.layers.concatenate([r1mu,r1logvar,r1w])
        self.encoder_r1 = tf.keras.Model(inputs=all_input_y, outputs=r1)
        print(self.encoder_r1.summary())

        # the q encoder network
        q_input_x = tf.keras.Input(shape=(self.x_dim))
        #qc = tf.keras.layers.Conv1D(filters=32, kernel_size=16, strides=1, activation=self.act)(all_input_y)
        #qc = tf.keras.layers.MaxPooling1D(pool_size=2)(qc)
        #qc = tf.keras.layers.Conv1D(filters=32, kernel_size=16, strides=4, activation=self.act)(qc)
        #qc = tf.keras.layers.MaxPooling1D(pool_size=2)(qc)
        #qc = tf.keras.layers.Conv1D(filters=32, kernel_size=16, strides=4, activation=self.act)(qc)
        #qc = tf.keras.layers.MaxPooling1D(pool_size=2)(qc) 
        #qc = tf.keras.layers.Flatten()(qc)
        qx = tf.keras.layers.Flatten()(q_input_x)
        q = tf.keras.layers.concatenate([r1c,qx])        
        q = tf.keras.layers.Dense(2048,activation=self.act)(q)
        q = tf.keras.layers.Dense(2048,activation=self.act)(q)
        q = tf.keras.layers.Dense(1024,activation=self.act)(q)
        q = tf.keras.layers.Dense(2*self.z_dim)(q)
        self.encoder_q = tf.keras.Model(inputs=[all_input_y, q_input_x], outputs=q)
        print(self.encoder_q.summary())

        # the r2 decoder network
        r2_input_z = tf.keras.Input(shape=(self.z_dim))
        #r2c = tf.keras.layers.Conv1D(filters=32, kernel_size=16, strides=1, activation=self.act)(all_input_y)
        #r2c = tf.keras.layers.MaxPooling1D(pool_size=2)(r2c)
        #r2c = tf.keras.layers.Conv1D(filters=32, kernel_size=16, strides=1, activation=self.act)(r2c)
        #r2c = tf.keras.layers.MaxPooling1D(pool_size=2)(r2c)
        #r2c = tf.keras.layers.Conv1D(filters=32, kernel_size=16, strides=1, activation=self.act)(r2c)
        #r2c = tf.keras.layers.MaxPooling1D(pool_size=2)(r2c)
        #r2c = tf.keras.layers.Flatten()(r2c)
        r2z = tf.keras.layers.Flatten()(r2_input_z)
        r2 = tf.keras.layers.concatenate([r1c,r2z])
        r2 = tf.keras.layers.Dense(1024,activation=self.act)(r2)
        r2 = tf.keras.layers.Dropout(0.2)(r2)
        r2 = tf.keras.layers.Dense(512,activation=self.act)(r2)
        r2 = tf.keras.layers.Dropout(0.2)(r2)
        r2 = tf.keras.layers.Dense(512,activation=self.act)(r2)
        r2 = tf.keras.layers.Dropout(0.2)(r2)
        r2 = tf.keras.layers.Dense(2*self.x_dim)(r2)
        self.decoder_r2 = tf.keras.Model(inputs=[all_input_y, r2_input_z], outputs=r2)
        #self.decoder_r2 = tf.keras.Model(inputs=r2_input_z, outputs=r2)
        print(self.decoder_r2.summary())

    def encode_r1(self, y=None):
        mean, logvar, weight = tf.split(self.encoder_r1(y), num_or_size_splits=[self.z_dim*self.n_modes, self.z_dim*self.n_modes,self.n_modes], axis=1)
        #weight = tf.reshape(weight,[-1,self.n_modes])
        return tf.reshape(mean,[-1,self.n_modes,self.z_dim]), tf.reshape(logvar,[-1,self.n_modes,self.z_dim]), tf.reshape(weight,[-1,self.n_modes])
        #return tf.reshape(mean,[-1,self.n_modes,self.z_dim]), tf.reshape(logvar,[-1,self.n_modes,self.z_dim]), tf.zeros(shape=weight.shape)

    def encode_q(self, x=None, y=None):
        return tf.split(self.encoder_q([y,x]), num_or_size_splits=[self.z_dim,self.z_dim], axis=1)

    def decode_r2(self, y=None, z=None, apply_sigmoid=False):
        mean, logvar = tf.split(self.decoder_r2([y,z]), num_or_size_splits=[self.x_dim, self.x_dim], axis=1)
        return tf.reshape(mean,[-1,self.x_dim]), tf.reshape(logvar,[-1,self.x_dim])

def compute_loss(model, x, y, ramp=1.0, noiseamp=1.0):

    # randomise distance    
    #old_d = bounds['luminosity_distance_min'] + tf.boolean_mask(x,dist_mask,axis=1)*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
    #new_x = tf.random.uniform(shape=tf.shape(old_d), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
    #new_d = bounds['luminosity_distance_min'] + new_x*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
    #x = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_dist_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_dist_mask),axis=1)
    #dist_scale = tf.tile(tf.expand_dims(old_d/new_d,axis=1),(1,tf.shape(y)[1],1))

    # augment the data for this chunk - randomize phase and arrival time
    #aold_phase = bounds['phase_min'] + tf.boolean_mask(x,phase_mask,axis=1)*(bounds['phase_max'] - bounds['phase_min'])
    #new_x = tf.random.uniform(shape=tf.shape(old_phase), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
    #new_phase = bounds['phase_min'] + new_x*(bounds['phase_max'] - bounds['phase_min'])
    #x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_phase_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_phase_mask),axis=1)
    #old_geocent = bounds['geocent_time_min'] + tf.boolean_mask(x,geocent_mask,axis=1)*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
    #new_x = tf.random.uniform(shape=tf.shape(old_geocent), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
    #new_geocent = bounds['geocent_time_min'] + new_x*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
    #x = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_geocent_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_geocent_mask),axis=1)
    #phase_correction = -1.0*tf.complex(tf.cos(new_phase-old_phase),tf.sin(new_phase-old_phase))
    #phase_correction = tf.tile(tf.expand_dims(phase_correction,axis=1),(1,3,tf.shape(y)[1]/2 + 1))
    #fvec = tf.range(params['ndata']/2 + 1)/params['duration']
    #time_correction = -2.0*np.pi*fvec*(new_geocent-old_geocent)
    #time_correction = tf.complex(tf.cos(time_correction),tf.sin(time_correction))
    #time_correction = tf.tile(tf.expand_dims(time_correction,axis=1),(1,3,1))
    #y_fft = tf.signal.rfft(tf.transpose(y,[0,2,1]))*phase_correction*time_correction
    #y = tf.transpose(tf.signal.irfft(y_fft),[0,2,1])

    # add noise and randomise distance again
    y = (y + noiseamp*tf.random.normal(shape=tf.shape(y), mean=0.0, stddev=1.0, dtype=tf.float32))/params['y_normscale']
    mean_r1, logvar_r1, logweight_r1 = model.encode_r1(y=y)
    scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
    gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r1),
            components_distribution=tfd.MultivariateNormalDiag(
            loc=mean_r1,
            scale_diag=scale_r1))
    mean_q, logvar_q = model.encode_q(x=x,y=y)
    scale_q = EPS + tf.sqrt(tf.exp(logvar_q))
    mvn_q = tfp.distributions.MultivariateNormalDiag(
                          loc=mean_q,
                          scale_diag=scale_q)
    #mvn_q = tfd.Normal(loc=mean_q,scale=scale_q)
    z_samp = mvn_q.sample()
    mean_r2, logvar_r2 = model.decode_r2(z=z_samp,y=y)
    scale_r2 = EPS + tf.sqrt(tf.exp(logvar_r2))

    # SIMPLE RECON LOSS
    mvn_r2 = tfp.distributions.MultivariateNormalDiag(
                          loc=mean_r2,
                          scale_diag=scale_r2)
    simple_cost_recon = -1.0*tf.reduce_mean(mvn_r2.log_prob(x))

    selfent_q = -1.0*tf.reduce_mean(mvn_q.entropy())
    log_r1_q = gm_r1.log_prob(z_samp)   # evaluate the log prob of r1 at the q samples
    cost_KL = selfent_q - tf.reduce_mean(log_r1_q)
    return simple_cost_recon, cost_KL

def gen_samples(model, y, ramp=1.0, nsamples=1000, max_samples=1000):

    y = y/params['y_normscale']
    y = tf.tile(y,(max_samples,1,1))
    samp_iterations = int(nsamples/max_samples)
    for i in range(samp_iterations):
        mean_r1, logvar_r1, logweight_r1 = model.encode_r1(y=y)
        scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
        gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r1),
            components_distribution=tfd.MultivariateNormalDiag(
            loc=mean_r1,
            scale_diag=scale_r1))
        z_samp = gm_r1.sample()
        mean_r2, logvar_r2 = model.decode_r2(z=z_samp,y=y)
        scale_r2 = EPS + tf.sqrt(tf.exp(logvar_r2))
        mvn_r2 = tfp.distributions.MultivariateNormalDiag(
            loc=mean_r2,
            scale_diag=scale_r2)
      
        if i==0:
            x_sample = mvn_r2.sample()
        else:
            x_sample = tf.concat([x_sample,mvn_r2.sample()],axis=0)
    return x_sample

def gen_mode_samples(model, y, ramp=1.0, nsamples=1000, max_samples=1000):

    y = y/params['y_normscale']
    y = tf.tile(y,(max_samples,1,1))
    samp_iterations = int(nsamples/max_samples)
    x_sample = []
    for k in range(min(4,params['n_modes'])):
        for i in range(samp_iterations):
            mean_r1, logvar_r1, logweight_r1 = model.encode_r1(y=y)
            scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
            idx = np.argsort(logweight_r1[0,:])[::-1]
            weights = np.zeros(params['n_modes'])
            weights[idx] = 1.0     
            gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(probs=tf.cast(weights,dtype=tf.float32)),
                components_distribution=tfd.MultivariateNormalDiag(
                loc=mean_r1,
                scale_diag=scale_r1))
            z_samp = gm_r1.sample()
            mean_r2, logvar_r2 = model.decode_r2(z=z_samp,y=y)
            scale_r2 = EPS + tf.sqrt(tf.exp(logvar_r2))
            mvn_r2 = tfp.distributions.MultivariateNormalDiag(
                loc=mean_r2,
                scale_diag=scale_r2)

            if i==0:
                temp_x_sample = mvn_r2.sample()
            else:
                temp_x_sample = tf.concat([x_sample,mvn_r2.sample()],axis=0)
        x_sample.append(temp_x_sample)
    return x_sample

def ramp_func(epoch,start,ramp_length, n_cycles):
    i = (epoch-start)/(2.0*ramp_length)
    if i<0:
        return 0.0
    if i>=n_cycles:
        return 1.0
    return min(1.0,2.0*np.remainder(i,1.0))

def gen_z_samples(model, x, y, nsamples=1000,max_samples=1000):

    y = y/params['y_normscale']
    y = tf.tile(y,(max_samples,1,1))
    x = tf.tile(x,(max_samples,1))
    samp_iterations = int(nsamples/max_samples)
    for i in range(samp_iterations):

        mean_r1_temp, logvar_r1, logweight_r1_temp = model.encode_r1(y=y)
        scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
        gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r1_temp),
            components_distribution=tfd.MultivariateNormalDiag(
            loc=mean_r1_temp,
            scale_diag=scale_r1))
        mean_q_temp, logvar_q = model.encode_q(x=x,y=y)
        scale_q = EPS + tf.sqrt(tf.exp(logvar_q))
        mvn_q = tfp.distributions.MultivariateNormalDiag(
                          loc=mean_q_temp,
                          scale_diag=scale_q)
        if i==0:
            z_samp_q = mvn_q.sample()
            z_samp_r1 = gm_r1.sample()
            logweight_r1 = logweight_r1_temp
            mean_r1 = mean_r1_temp
            mean_q = mean_q_temp
        else:
            z_samp_q = tf.concat([z_samp_q,mvn_q.sample()],axis=0)
            z_samp_r1 = tf.concat([z_samp_r1,gm_r1.sample()],axis=0)
            logweight_r1 = tf.concat([logweight_r1,logweight_r1_temp],axis=0)
            mean_r1 = tf.concat([mean_r1,mean_r1_temp],axis=0)
            mean_q = tf.concat([mean_q,mean_q_temp],axis=0)
    
    return mean_r1, logweight_r1, z_samp_r1, mean_q, z_samp_q

@tf.function
def train_step(model, x, y, optimizer, ramp=1.0):
    """Executes one training step and returns the loss.
    This function computes the loss and gradients, and uses the latter to
    update the model's parameters.
    """
    with tf.GradientTape() as tape:
        r_loss, kl_loss = compute_loss(model, x, y, ramp=ramp)
        loss = r_loss + ramp*kl_loss
    gradients = tape.gradient(loss, model.trainable_variables)
    optimizer.apply_gradients(zip(gradients, model.trainable_variables))
    train_loss_metric(loss)
    return r_loss, kl_loss

epochs = 50000
train_loss_metric = tf.keras.metrics.Mean('train_loss', dtype=tf.float32)
train_log_dir = '/data/www.astro/chrism/vitamin_c/' + run + '/logs'
#os.mkdir(train_log_dir)
train_summary_writer = tf.summary.create_file_writer(train_log_dir)
model = CVAE(x_data_train.shape[1], params['ndata'], y_data_train.shape[2], params['z_dimension'], params['n_modes'])
optimizer = tf.keras.optimizers.Adam(1e-4)

# start the training loop
train_loss = np.zeros((epochs,3))
val_loss = np.zeros((epochs,3))
ramp_start = 100
#ramp_stop = 200
ramp_length = 100
#ramp_grad = 1.0/(ramp_stop - ramp_start)
ramp_cycles = 1
#boundary_start = 10
#boundary_stop = 20
#boundary_grad = 1.0/(boundary_stop - boundary_start)
KL_samples = []
for epoch in range(1, epochs + 1):

    train_loss_kl_q = 0.0
    train_loss_kl_r1 = 0.0
    start_time_train = time.time()
    ramp = tf.convert_to_tensor(ramp_func(epoch,ramp_start,ramp_length,ramp_cycles), dtype=tf.float32)
    #boundary = tf.convert_to_tensor(np.min([(epoch-boundary_start)*boundary_grad,1.0]).astype(np.single) if epoch>boundary_start else 0.0, dtype=tf.float32)
    for step, (x_batch_train, y_batch_train) in train_dataset.enumerate():
    #for step in range(training_generator.__len__()):
        #x_batch_train, y_batch_train = training_generator.__getitem__(step)
        temp_train_r_loss, temp_train_kl_loss = train_step(model, x_batch_train, y_batch_train, optimizer, ramp=ramp)
        train_loss[epoch-1,0] += temp_train_r_loss
        train_loss[epoch-1,1] += temp_train_kl_loss
    #training_generator.on_epoch_end()
    train_loss[epoch-1,2] = train_loss[epoch-1,0] + ramp*train_loss[epoch-1,1]
    train_loss[epoch-1,:] /= float(step+1)
    end_time_train = time.time()
    with train_summary_writer.as_default():
        tf.summary.scalar('loss', train_loss_metric.result(), step=epoch)
    train_loss_metric.reset_states()

    start_time_val = time.time()
    for step, (x_batch_val, y_batch_val) in val_dataset.enumerate():
        temp_val_r_loss, temp_val_kl_loss = compute_loss(model, x_batch_val, y_batch_val, ramp=ramp)
        val_loss[epoch-1,0] += temp_val_r_loss
        val_loss[epoch-1,1] += temp_val_kl_loss
    val_loss[epoch-1,2] = val_loss[epoch-1,0] + ramp*val_loss[epoch-1,1]
    val_loss[epoch-1,:] /= float(step+1)
    end_time_val = time.time()

    print('Epoch: {}, Run {}, Training RECON: {}, KL: {}, TOTAL: {}, time elapsed: {}'
        .format(epoch, run, train_loss[epoch-1,0], train_loss[epoch-1,1], train_loss[epoch-1,2], end_time_train - start_time_train))
    print('Epoch: {}, Run {}, Validation RECON: {}, KL: {}, TOTAL: {}, time elapsed {}'
        .format(epoch, run, val_loss[epoch-1,0], val_loss[epoch-1,1], val_loss[epoch-1,2], end_time_val - start_time_val))

    # update loss plot
    plot_losses(train_loss, val_loss, epoch, run=run)

    # generate and plot posterior samples for the latent space and the parameter space 
    if epoch % plot_cadence == 0:
        for step, (x_batch_test, y_batch_test, snrs_batch_test) in test_dataset.enumerate():             
            mu_r1, logw_r1, z_r1, mu_q, z_q = gen_z_samples(model, x_batch_test, y_batch_test, nsamples=8000)
            plot_latent(mu_r1,logw_r1,z_r1,mu_q,z_q,epoch,step,run=run)
            start_time_test = time.time()
            samples = gen_samples(model, y_batch_test, ramp=ramp, nsamples=8000)
            end_time_test = time.time()
            if np.any(np.isnan(samples)):
                print('Epoch: {}, found nans in samples. Not making plots'.format(epoch))
                for k,s in enumerate(samples):
                    if np.any(np.isnan(s)):
                        print(k,s)
                KL_est = [-1,-1,-1]
            else:
                print('Epoch: {}, run {} Testing time elapsed for 8000 samples: {}'.format(epoch,run,end_time_test - start_time_test))
                KL_est = plot_posterior(samples,x_batch_test[0,:],snrs_batch_test[0,:],epoch,step,all_other_samples=bilby_samples[:,step,:],run=run)
                _ = plot_posterior(samples,x_batch_test[0,:],snrs_batch_test[0,:],epoch,step,run=run)
            KL_samples.append(KL_est)
            mode_samples = gen_mode_samples(model,y_batch_test, ramp=ramp, nsamples=1000)
            plot_mode_posterior(mode_samples,x_batch_test[0,:],snrs_batch_test[0,:],epoch,step,run=run)

        # make p-p plot
        results = []
        for step, (x_batch_pp, y_batch_pp, snrs_batch_pp) in pp_dataset.enumerate():   #for each datafile, cast the true values and generated samples into an dictionary
            posterior = dict()
            injections = dict()
            priors = dict()
            samples = gen_samples(model,y_batch_pp,ramp=ramp,nsamples=1000)
            x_batch_pp = x_batch_pp.numpy().astype('float64')
            for i,x in enumerate(x_batch_pp[0]):
                posterior[str(i)] = samples[:,i]
                injections[str(i)] = x
                priors[str(i)] = Uniform(-100,100)
            posterior = pd.DataFrame(dict(posterior))
            result = bilby.result.Result(
                label="test",
                injection_parameters=injections,
                posterior=posterior,
                search_parameter_keys=injections.keys(),
                priors=priors)
            results.append(result)
        bilby.result.make_pp_plot(results, filename='/data/www.astro/chrism/vitamin_c/%s/pp_plot_epoch_%d.png' % (run,epoch),
                              confidence_interval=[0.68,0.95,0.997])

        # plot KL evolution
        #plot_KL(np.reshape(np.array(KL_samples),[-1,4,3]),plot_cadence,run=run)

    # load more noisefree training data back in
    if epoch % 4 == 0:
        x_data_train, y_data_train, _, snrs_train = load_data(params,bounds,fixed_vals,params['train_set_dir'],params['inf_pars'],silent=True)

        # augment the data for this chunk - distqnce
        old_d = bounds['luminosity_distance_min'] + tf.boolean_mask(x_data_train,dist_mask,axis=1)*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
        new_x = tf.random.uniform(shape=tf.shape(old_d), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
        new_d = bounds['luminosity_distance_min'] + new_x*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
        x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_dist_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_dist_mask),axis=1)
        dist_scale = tf.tile(tf.expand_dims(old_d/new_d,axis=1),(1,tf.shape(y_data_train)[1],1))

        # augment the data for this chunk - randomize phase and arrival time
        old_phase = bounds['phase_min'] + tf.boolean_mask(x_data_train,phase_mask,axis=1)*(bounds['phase_max'] - bounds['phase_min'])
        new_x = tf.random.uniform(shape=tf.shape(old_phase), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
        new_phase = bounds['phase_min'] + new_x*(bounds['phase_max'] - bounds['phase_min'])
        x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_phase_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_phase_mask),axis=1)
        old_geocent = bounds['geocent_time_min'] + tf.boolean_mask(x_data_train,geocent_mask,axis=1)*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
        new_x = tf.random.uniform(shape=tf.shape(old_geocent), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
        new_geocent = bounds['geocent_time_min'] + new_x*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
        x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_geocent_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_geocent_mask),axis=1)
        phase_correction = -1.0*tf.complex(tf.cos(new_phase-old_phase),tf.sin(new_phase-old_phase))
        phase_correction = tf.tile(tf.expand_dims(phase_correction,axis=1),(1,3,tf.shape(y_data_train)[1]/2 + 1))
        fvec = tf.range(params['ndata']/2 + 1)/params['duration']
        time_correction = -2.0*np.pi*fvec*(new_geocent-old_geocent)
        time_correction = tf.complex(tf.cos(time_correction),tf.sin(time_correction))
        time_correction = tf.tile(tf.expand_dims(time_correction,axis=1),(1,3,1))
        y_data_train_fft = tf.signal.rfft(tf.transpose(y_data_train,[0,2,1]))*phase_correction*time_correction
        y_data_train = tf.transpose(tf.signal.irfft(y_data_train_fft),[0,2,1])*dist_scale

        train_dataset = (tf.data.Dataset.from_tensor_slices((x_data_train,y_data_train))
                 .shuffle(train_size).batch(batch_size))





















